@@ -41,7 +41,7 @@
 #define DRIVER_LOGO	PRODUCT_LOGO " driver"
 #define SHORT_LOGO	"fcdsl-" INTERFACE
 
-#define	REV_DEFAULT	"0.2"
+#define	REV_DEFAULT	"0.3"
 
 #define PCI_DEVICE_ID_TM                0x5402
 #define PCI_DEVICE_ID_AVM_FCDSL         0x0f00
@@ -56,7 +56,13 @@
 # define LOG_LINK
 #endif
 
-#define LOG_TIMER
+/*---------------------------------------------------------------------------*\
+ * LOG_MESSAGES - CAPI trace
+ * LOG_ALLOC - Log heap activity
+ * LOG_TIMER - Activate debug timers
+\*---------------------------------------------------------------------------*/
+
+#define	LOG_MESSAGES
 
 #define	UNUSED_ARG(x)	(x)=(x)
 
