@@ -99,9 +99,6 @@ struct capi_driver_interface *	capi_driver		= NULL;
 static int			load_flag		= FALSE;
 static int			init_flag		= FALSE;
 static int			nvers			= 0;
-static struct tq_struct		tq_dpc;			/* for tq_immediate */
-static struct tq_struct		tq_tx_dpc;		/* for tq_private */
-static struct tq_struct		tq_rx_dpc;		/* for tq_private */
 static atomic_t			link_open		= ATOMIC_INIT (0);
 static atomic_t			scheduler_enabled	= ATOMIC_INIT (1);
 static atomic_t			tx_pending		= ATOMIC_INIT (0);
@@ -110,21 +107,23 @@ static atomic_t			crit_count		= ATOMIC_INIT (0);
 struct capi_ctr *		capi_controller[2]	= { NULL, NULL } ;
 static bundle_t			ctrl_context[2];
 static per_ctrl_t		ctrl_params[2];
-static unsigned long		qtflags;
-static spinlock_t		qtlock			= SPIN_LOCK_UNLOCKED;
-static atomic_t			qtlockcount 		= ATOMIC_INIT (0);
+static spinlock_t		stack_lock		= SPIN_LOCK_UNLOCKED;
+static atomic_t			rx_flag			= ATOMIC_INIT (0);
+static atomic_t			tx_flag			= ATOMIC_INIT (0);
+static unsigned long		qt_flags;
+static spinlock_t		qt_lock			= SPIN_LOCK_UNLOCKED;
 static atomic_t			thread_flag		= ATOMIC_INIT (0);
 static atomic_t			thread_capi_flag	= ATOMIC_INIT (0);
 static int			thread_pid		= -1;
 
-static DECLARE_TASK_QUEUE(tq_private);
 static DECLARE_WAIT_QUEUE_HEAD(wait);
 static DECLARE_WAIT_QUEUE_HEAD(capi_wait);
 static DECLARE_MUTEX_LOCKED(hotplug);
 
-static void			dpc (void *);
-static void			tx_dpc (void *);
-static void			rx_dpc (void *);
+static void                     tx_task (unsigned long data);
+static void                     rx_task (unsigned long data);
+static void			tx_handler (card_p cp);
+static void			rx_handler (card_p cp);
 static void			kick_scheduler (void);
 static void			irq_handler (int, void *, struct pt_regs *);
 static void			enable_thread (void);
@@ -139,6 +138,9 @@ static unsigned			scheduler_resume (void);
 static unsigned			controller_remove (void);
 static unsigned			controller_add (void);
 
+static DECLARE_TASKLET(tx_tasklet, tx_task, 0);
+static DECLARE_TASKLET(rx_tasklet, rx_task, 0);
+
 struct capi_driver		capi_interface		= {
 
 	TARGET,
@@ -278,15 +280,15 @@ static void stop_card (card_p cp) {
 	POKE (cp->io_base + ICLEAR, ~0U);
 	/* Reset CPU */
 	val = PEEK (cp->io_base + BIU_CTL);
-	log (__FUNCTION__ "(%d): BIU_CTL = 0x%08x\n", __LINE__, val);
+	log ("stop: BIU_CTL = 0x%08x\n", val);
 	
 	POKE (cp->io_base + BIU_CTL, val & ~CARD_BIU_CTL_CR);
 	val = PEEK (cp->io_base + BIU_CTL);
-	log (__FUNCTION__ "(%d): BIU_CTL = 0x%08x\n", __LINE__, val);
+	log ("stop: BIU_CTL = 0x%08x\n", val);
 	
 	POKE (cp->io_base + BIU_CTL, val | CARD_BIU_CTL_SR);
 	val = PEEK (cp->io_base + BIU_CTL);
-	log (__FUNCTION__ "(%d): BIU_CTL = 0x%08x\n", __LINE__, val);
+	log ("stop: BIU_CTL = 0x%08x\n", val);
 } /* stop_card */
 
 /*---------------------------------------------------------------------------*\
@@ -294,7 +296,7 @@ static void stop_card (card_p cp) {
 static int reset_card (card_p cp) {
 	unsigned val, N;
 
-	log (__FUNCTION__ "(%d): BIU_CTL = 0x%08x\n", __LINE__, PEEK (cp->io_base + BIU_CTL));
+	log ("reset: BIU_CTL = 0x%08x\n", PEEK (cp->io_base + BIU_CTL));
 	/* Reset CPU */
 	stop_card (cp);
 	
@@ -304,10 +306,10 @@ static int reset_card (card_p cp) {
 	/* Reset DMA */
 	N = 0;
 	while (1 == (PEEK (cp->io_base + BIU_STATUS) & (1<<4))) {	
-		log (__FUNCTION__ ": Waiting for DMA completion...\n");
+		log ("Waiting for DMA completion...\n");
 		mdelay (100);
 		if (++N == DMA_RESET_LOOPS) {
-			log (__FUNCTION__ ": Waited too long.\n");
+			log ("Waited too long.\n");
 			return FALSE;
 		}
 	}
@@ -362,14 +364,14 @@ static int reset_card (card_p cp) {
 	POKE (cp->io_base + JTAG_CTL, 4);
 
 	/* Reset peripheral chip */
-	log (__FUNCTION__ ": Peripheral chip reset.\n");
+	log ("Peripheral chip reset.\n");
 	POKE (cp->io_base + INT_CTL, 1<<2 | 1<<6);
 
 	/* Block Power Down */
-	log (__FUNCTION__ ": Block power down.\n");
+	log ("Block power down.\n");
 	POKE (cp->io_base + BLOCK_POWER_DOWN, ~0U);
 	
-	log (__FUNCTION__ " completed.\n");
+	log ("Reset completed.\n");
 	return TRUE;
 } /* reset_card */
 
@@ -391,7 +393,7 @@ static int test_card (card_p cp) {
 	/* SDRAM Test */
 	DOTEST (mem_base, SDRAM_FIRMWARE_BASE);
 	
-	log (__FUNCTION__ ": completed.\n");
+	log ("Test completed.\n");
 	return TRUE;
 	
 #undef DOTEST
@@ -476,7 +478,7 @@ static int find_card (card_p cp) {
 static struct capi_ctr * get_capi_ctr (int ctrl) {
 	int	ix;
 	
-	assert (capi_card);
+	assert (capi_card != NULL);
 	assert (capi_card->ctrl2 > 0);
 	ix = (capi_card->ctrl2 == ctrl) ? 1 : 0;
 	info (NULL != capi_controller[ix]);
@@ -490,7 +492,7 @@ static int nbchans (struct capi_ctr * ctrl) {
 	unsigned char *	prf;
 	int		temp = 2;
     
-	assert (ctrl);
+	assert (ctrl != NULL);
 	assert (ctrl->driverdata);
 	C = GET_CTRL (ctrl);
 	prf = (unsigned char *) C->string[6];
@@ -527,10 +529,10 @@ static void scan_version (struct capi_ctr * ctrl, const char * ver) {
 	char *		vstr;
 	per_ctrl_p	C;
 
-	assert (ctrl);
+	assert (ctrl != NULL);
 	assert (ctrl->driverdata);
 	C = GET_CTRL (ctrl);
-
+	assert (C != NULL);
 	vlen = (unsigned char) ver[0];
 	C->version = vstr = (char *) hmalloc (vlen);
 	if (NULL == vstr) {
@@ -558,10 +560,10 @@ static void copy_version (struct capi_ctr * ctrl) {
 	char *		tmp;
 	per_ctrl_p	C;
 	
-	assert (ctrl);
+	assert (ctrl != NULL);
 	assert (ctrl->driverdata);
 	C = GET_CTRL (ctrl);
-	
+	assert (C != NULL);
 	if (NULL == (tmp = C->string[3])) {
 		lprintf (KERN_ERR, "Do not have version information...\n");
 		return;
@@ -585,10 +587,10 @@ static void kill_version (struct capi_ctr * ctrl) {
 	int		i;
 	per_ctrl_p	C;
 	
-	assert (ctrl);
+	assert (ctrl != NULL);
 	assert (ctrl->driverdata);
 	C = GET_CTRL (ctrl);
-
+	assert (C != NULL);
 	for (i = 0; i < 8; i++) {
 		C->string[i] = NULL;
 	}
@@ -615,7 +617,6 @@ static int start_stack (card_p card) {
 	card->count = 0;
 	table_init (&card->appls);
 	queue_init (&card->queue);
-	enable_thread ();
 	(*capi_lib->cm_register_ca_functions) (&ca_funcs);
 	if ((*capi_lib->cm_start) ()) {
 		log ("Starting the stack failed...\n");
@@ -634,6 +635,7 @@ static int start_stack (card_p card) {
 	}
 	/* (*capi_lib->cm_handle_events) (); */
 	leave_critical ();
+	enable_thread ();
 	init_flag = 1;
 	return TRUE; 
 } /* start_stack */
@@ -646,17 +648,17 @@ static void stop_stack (card_p card) {
 		log ("Stack not initialized.\n");
 		return;
 	}
-	(*capi_lib->cm_exit) ();
-	mdelay (200);
 	if (thread_pid != -1) {
 		disable_thread ();
 	}
+	(*capi_lib->cm_exit) ();
 	queue_exit (&card->queue);
 	table_exit (&card->appls);
 	if (ctrl_context[0].ctrl != NULL) {
 		kill_version (capi_controller[0]);
 	}
 	kill_version (capi_controller[1]);
+	nvers = 0;
 } /* stop_stack */
 
 /*---------------------------------------------------------------------------*\
@@ -701,12 +703,12 @@ static int relocate_and_load (
 
 	size = align = -1;
 	CHECK( lib_getsize (fw, &size, &align) );
-	log (__FUNCTION__ "(%d) : size = %d, align = %d\n", __LINE__, size, align);
+	log ("load: size = %d, align = %d\n", size, align);
 	assert ((unsigned) size <= cp->len_sdram);
 	
 	usize = IMAGE_ADD_ROUND (size, /* add */ 1<<15, /* round */ 1<<6);
 	assert (usize < cp->len_sdram);
-	log (__FUNCTION__ "(%d) : rounded size = %u\n", __LINE__, usize);
+	log ("load: rounded size = %u\n", usize);
 	
 	CHECK( lib_getimage (fw, IMAGE_FREQUENCY, usize) );
 	CHECK( lib_load_symtab (fw) );
@@ -726,8 +728,8 @@ static int relocate_and_load (
 	
 error_exit:
 	lib_remove (&fw);
-	log (__FUNCTION__ "(%d) : %s FAILED!\n", chk_line, chk_stmt); 
-	log (__FUNCTION__ "(%d) : reason \"%s\"[%d].\n", lib_error (res), res);
+	log ("load: In line %d: %s FAILED!\n", chk_line, chk_stmt); 
+	log ("load: reason \"%s\"[%d].\n", lib_error (res), res);
 	return res;
 
 #undef CHECK
@@ -745,15 +747,15 @@ static void start_ware (card_p cp) {
 	POKE (cp->io_base + ICLEAR, ~0U);
 
 	val = PEEK (cp->io_base + BIU_CTL);
-	log (__FUNCTION__ "(%d) : 0: 0x%08x\n", __LINE__, val);
+	log ("start: 0: 0x%08x\n", val);
 
 	POKE (cp->io_base + BIU_CTL, val & ~((unsigned) CARD_BIU_CTL_SR));
 	val = PEEK (cp->io_base + BIU_CTL);
-	log (__FUNCTION__ "(%d) : 1: 0x%08x\n", __LINE__, val);
+	log ("start: 1: 0x%08x\n", val);
 	
 	POKE (cp->io_base + BIU_CTL, val |  ((unsigned) CARD_BIU_CTL_CR));
 	val = PEEK (cp->io_base + BIU_CTL);
-	log (__FUNCTION__ "(%d) : 2: 0x%08x\n", __LINE__, val);
+	log ("start: 2: 0x%08x\n", val);
 } /* start_ware */
 
 /*---------------------------------------------------------------------------*\
@@ -804,7 +806,7 @@ void * map_addr_from_tm_ptr (card_p cp, unsigned long X) {
 	res -= cp->addr_sdram;
 	assert (res < cp->len_sdram);
 	res += (unsigned) cp->mem_base;
-	log (__FUNCTION__ "() : TM 0x%08x mapped to 0x%08x\n", X, res);
+	log ("map: TM 0x%08x mapped to 0x%08x\n", X, res);
 	return (void *) res;
 } /* deref_map_addr_from_tm_ptr */
 
@@ -814,7 +816,7 @@ void * deref_map_addr_from_tm_ptr (card_p cp, unsigned long X) {
 	unsigned long	res;
 	
 	res = PEEK (cp->mem_base + (X - (unsigned long) cp->addr_sdram));
-	log (__FUNCTION__ "() : TM 0x%08x = %08x\n", X, res);
+	log ("deref: TM 0x%08x = %08x\n", X, res);
 	return map_addr_from_tm_ptr (cp, res);
 } /* deref_map_addr_from_tm_ptr */
 
@@ -1049,26 +1051,6 @@ int __entry add_card (struct capi_driver * drv, capicardparams * args) {
 	cp->ctxp  = NULL;
 	cp->rx_buffer = cp->tx_buffer = NULL;
 	cp->rx_func   = cp->tx_func   = NULL;
-
-#if LINUX_VERSION_CODE < KERNEL_VERSION(2,4,0)
-	tq_dpc.next = 
-	tq_tx_dpc.next = 
-	tq_rx_dpc.next = NULL;
-#else
-	tq_dpc.list.next = 
-	tq_tx_dpc.list.next = 
-	tq_rx_dpc.list.next = NULL;
-#endif
-	tq_dpc.sync = 
-	tq_tx_dpc.sync =
-	tq_rx_dpc.sync = 0;
-	tq_dpc.data = 
-	tq_tx_dpc.data =
-	tq_rx_dpc.data = cp;
-	tq_dpc.routine = &dpc;
-	tq_tx_dpc.routine = &tx_dpc;
-	tq_rx_dpc.routine = &rx_dpc;
-						
 	return 0;
 } /* add_card */ 
 
@@ -1102,52 +1084,63 @@ void __entry remove_ctrl (struct capi_ctr * ctrl) {
 void lock (void) {
 	unsigned long	local_flags;
 	
-	if (0 == atomic_inc_and_test (&qtlockcount)) {
-		spin_lock_irqsave (&qtlock, local_flags);
-		qtflags = local_flags;
-	}
+	info (!spin_is_locked (&qt_lock));
+	spin_lock_irqsave (&qt_lock, local_flags);
+	qt_flags = local_flags;
 } /* lock */
 
 /*---------------------------------------------------------------------------*\
 \*---------------------------------------------------------------------------*/
 void unlock (void) {
-	unsigned long	local_flags = qtflags;
+	unsigned long	local_flags = qt_flags;
 
-	assert (atomic_read (&qtlockcount) > 0);
-	if (1 == atomic_dec_and_test (&qtlockcount)) {
-		spin_unlock_irqrestore (&qtlock, local_flags);
-	}
+	spin_unlock_irqrestore (&qt_lock, local_flags);
 } /* unlock */
 
 /*---------------------------------------------------------------------------*\
 \*-C-------------------------------------------------------------------------*/
-void enter_critical (void) {
-   
-	atomic_inc (&crit_count);
-} /* enter_critical */
-
-/*---------------------------------------------------------------------------*\
-\*---------------------------------------------------------------------------*/
 static inline int in_critical (void) {
 
 	return (0 < atomic_read (&crit_count));
 } /* in_critical */
 
+/*---------------------------------------------------------------------------*\
+\*---------------------------------------------------------------------------*/
+static inline void check (void) {
+	unsigned long	flags;
+	spinlock_t	chk_lock = SPIN_LOCK_UNLOCKED;
+
+	if (atomic_xchg (&rx_flag, 0)) {
+		spin_lock_irqsave (&chk_lock, flags);
+		rx_handler (capi_card);
+		spin_unlock_irqrestore (&chk_lock, flags);
+	}
+	if (atomic_xchg (&tx_flag, 0)) {
+		spin_lock_irqsave (&chk_lock, flags);
+		tx_handler (capi_card);
+		spin_unlock_irqrestore (&chk_lock, flags);
+	}
+} /* check */
+
+/*---------------------------------------------------------------------------*\
+\*---------------------------------------------------------------------------*/
+void enter_critical (void) {
+
+	if (!atomic_read (&crit_count) && !in_interrupt ()) {
+		check ();
+	}
+	atomic_inc (&crit_count);
+} /* enter_critical */
+
 /*---------------------------------------------------------------------------*\
 \*---------------------------------------------------------------------------*/
 void leave_critical (void) {
-	static spinlock_t	lock = SPIN_LOCK_UNLOCKED;
-	unsigned long   	flags;
 	
 	assert (in_critical ());
-	spin_lock_irqsave (&lock, flags);
-	if (1 == atomic_read (&crit_count)) {	
-		/* About to leave critical area... */
-		queue_task (&tq_dpc, &tq_immediate); 
-		mark_bh (IMMEDIATE_BH);	
-	}
 	atomic_dec (&crit_count);
-	spin_unlock_irqrestore (&lock, flags);
+	if (!atomic_read (&crit_count) && !in_interrupt ()) {
+		check ();
+	}
 } /* leave_critical */
 
 /*---------------------------------------------------------------------------*\
@@ -1191,6 +1184,7 @@ void __entry register_appl (
 		if (NULL == ptr) {
 			lprintf (KERN_ERR, "Not enough memory for application data.\n");
 			remove_appl (card->appls, appp);
+			return;
 		} else {
 			inc_use_count ();
 			card->count++;
@@ -1213,14 +1207,14 @@ void __entry release_appl (struct capi_ctr * ctrl, u16 appl) {
 	mlog ("RELEASE(ctrl:%d,appl:%u)\n", ctrl->cnr, appl);
 	assert (ctrl);
 	assert (ctrl->driverdata);
-	card =  GET_CARD (ctrl);
+	card = GET_CARD (ctrl);
 	if (ctrl->cnr == card->ctrl2) {	
 		if (NULL == (appp = search_appl (card->appls, appl))) {
 			log ("Attempt to release unknown application (id #%u)\n", appl);
 			return;
 		}
 		skb = make_0xfe_request (appl);	
-		handle_message (card->queue, appp, skb);
+		handle_message (card->appls, appp, skb);
 		kick_scheduler ();
 		appp->dying = TRUE;
 	}
@@ -1245,7 +1239,7 @@ void __entry send_msg (struct capi_ctr * ctrl, struct sk_buff * skb) {
 		log ("Message for unknown application (id %u).\n", appl);
 		return;
 	}
-	handle_message (card->queue, appp, skb);
+	handle_message (card->appls, appp, skb);
 	kick_scheduler ();
 } /* send_msg */
 
@@ -1459,11 +1453,11 @@ int msg2stack (unsigned char * msg) {
 			*dptr++ = temp & 0xFF;	temp >>= 8;
 			*dptr++ = temp & 0xFF;	temp >>= 8;
 			*dptr++ = temp & 0xFF;
+			memcpy (msg, mptr, mlen);
 			queue_park (capi_card->queue, appl, ncci, hand);
-			memcpy (msg, mptr, mlen);
 		} else {
-			queue_drop (capi_card->queue);
 			memcpy (msg, mptr, mlen);
+			queue_drop (capi_card->queue);
 		}
 	}
 	return res;
@@ -1487,13 +1481,15 @@ void msg2capi (unsigned char * msg) {
 	mlen = msg[0] + 256 * msg[1];
 	appl = msg[2] + 256 * msg[3];
 	nctr = msg[8] & 127;
+	
 	mlog (
-		"GET_MESSAGE(ctrl:%d,appl:%d,cmd:0x%02X,subcmd:0x%02X)\n", 
+		"GET_MESSAGE(ctrl:%u,appl:%d,cmd:0x%02X,subcmd:0x%02X)\n", 
 		nctr,
 		appl, 
 		msg[4], 
 		msg[5]
 	);
+	
 	if ((0x86 == msg[4]) && (0x81 == msg[5])) {	/* DATA_B3_CONF */
 		hand = msg[12] + 256 * msg[13];
 		ncci = msg[8] 
@@ -1503,7 +1499,6 @@ void msg2capi (unsigned char * msg) {
 				)	);
 		queue_conf (capi_card->queue, appl, ncci, hand);
 	}
-
 	if ((0x86 == msg[4]) && (0x82 == msg[5])) {	/* DATA_B3_IND */
 		dlen = msg[16] + 256 * msg[17];
 		skb = alloc_skb (
@@ -1551,6 +1546,7 @@ void msg2capi (unsigned char * msg) {
 	info (ctrl != NULL);
 	if (ctrl != NULL) {
 		assert (ctrl->cnr == (int) nctr);
+		assert (ctrl->handle_capimsg != NULL);
 		(*ctrl->handle_capimsg) (ctrl, appl, skb);
 	}
 } /* msg2capi */
@@ -1569,7 +1565,7 @@ void new_ncci (
 	unsigned		nctr;
 	
 	nctr = ncci & 127;
-	mlog ("NEW NCCI(ctrl:%d,appl:%u,ncci:0x%08X)\n", nctr, appl_id, ncci);
+	mlog ("NEW NCCI(ctrl:%u,appl:%u,ncci:0x%08X)\n", nctr, appl_id, ncci);
 	if (NULL == (appp = search_appl (capi_card->appls, appl_id))) {
 		lprintf (KERN_ERR, invalid_msg, appl_id);
 		return;
@@ -1586,8 +1582,9 @@ void new_ncci (
 		return;
 	}
 	ctrl = get_capi_ctr (nctr);
-	assert (ctrl);	
+	assert (ctrl != NULL);	
 	assert (ctrl->cnr == (int) nctr);
+	assert (ctrl->new_ncci != NULL);
 	(*ctrl->new_ncci) (ctrl, appl_id, ncci, winsize);
 } /* new_ncci */
 
@@ -1622,7 +1619,7 @@ void free_ncci (unsigned appl_id, __u32 ncci) {
 		ctrl = get_capi_ctr (nctr);
 		assert (ctrl);		
 		assert (ctrl->cnr == (int) nctr);
-		mlog ("FREE NCCI(ctrl:%d, appl:%u,ncci:0x%08X)\n", nctr,  appl_id, ncci);
+		mlog ("FREE NCCI(ctrl:%u, appl:%u,ncci:0x%08X)\n", nctr,  appl_id, ncci);
 		(*ctrl->free_ncci) (ctrl, appl_id, ncci);
 		remove_ncci (capi_card->appls, appp, nccip);
 	} else {
@@ -1645,12 +1642,31 @@ unsigned char * data_block (unsigned appl_id, __u32 ncci, unsigned handle) {
 
 /*-S-------------------------------------------------------------------------*\
 \*---------------------------------------------------------------------------*/
+static void scheduler_control (unsigned ena) {
+	int	enabled = (int) ena;
+	int	changed;
+
+	enter_critical ();
+	changed = (atomic_read (&scheduler_enabled) != enabled);
+	leave_critical ();
+	if (changed) {
+		if (enabled) {
+			kick_scheduler ();
+		} else {
+			atomic_set (&scheduler_enabled, 0);
+			wake_up_interruptible (&wait);
+		}
+	}
+} /* scheduler_control */
+
+/*---------------------------------------------------------------------------*\
+\*---------------------------------------------------------------------------*/ 
 static int sched_thread (void * arg) {
 
 	UNUSED_ARG (arg);
 	daemonize ();
 	SNPRINTF (current->comm, 16, "%s_thread", TARGET);	/* See: sched.h */
-	log ("Starting kicker thread '%s'...\n", current->comm);
+	log ("Starting scheduler thread '%s'...\n", current->comm);
 	while (atomic_read (&thread_flag)) {
 		/* Start/stop logic */
 		if (!atomic_read (&thread_capi_flag)) {
@@ -1662,7 +1678,6 @@ static int sched_thread (void * arg) {
 				interruptible_sleep_on (&wait);
 			}
 		}
-		assert (!in_critical ());
 		/* Enable/disable logic */
 		if (!atomic_read (&thread_flag)) {
 			info (atomic_read (&thread_flag));
@@ -1677,10 +1692,13 @@ static int sched_thread (void * arg) {
 			continue;
 		}
 		/* Body of thread, invoke scheduler */
-		os_timer_poll ();
-		assert (capi_lib->cm_schedule);
-		if ((*capi_lib->cm_schedule) ()) {
-			scheduler_control (TRUE); 
+		if (spin_trylock (&stack_lock)) {
+			os_timer_poll ();
+			assert (capi_lib->cm_schedule);
+			if ((*capi_lib->cm_schedule) ()) {
+				scheduler_control (TRUE); 
+			}
+			spin_unlock (&stack_lock);
 		}
 	}
 	log ("Scheduler thread stopped.\n");
@@ -1745,19 +1763,8 @@ static void disable_thread (void) {
 
 /*---------------------------------------------------------------------------*\
 \*---------------------------------------------------------------------------*/
-static void dpc (void * data) {
-	
-	UNUSED_ARG (data);
-	if (!in_critical ()) {
-		run_task_queue (&tq_private);
-	}
-} /* dpc */
-
-/*---------------------------------------------------------------------------*\
-\*---------------------------------------------------------------------------*/
-static void tx_dpc (void * data) {
+static void tx_handler (card_p cp) {
 	int			is_open;
-	card_p			cp;
 	completer_func_t	fp;
 	void *			bufp;
 	int			pending;
@@ -1765,17 +1772,15 @@ static void tx_dpc (void * data) {
 	xfer_info_t		info;
 #endif
 	
+	atomic_set (&tx_flag, 0);
 	is_open = atomic_read (&link_open);
 	info (is_open);
 	if (!is_open) {
 		return;
 	}
-	
 	pending = atomic_xchg (&tx_pending, 1);
 	assert (pending == 0);
-	
-	cp = (card_p) data;
-	assert (cp);
+	assert (cp != NULL);
 #if !defined (NDEBUG)
 	assert (PEEK (cp->io_base + XFER_TOTM_STATUS) == PC_PENDING);
 	xfer_info_get (tx_info, 0, &info);
@@ -1791,15 +1796,30 @@ static void tx_dpc (void * data) {
 		(*fp) (bufp, cp->tx_length, cp->tx_ctx);
 		kick_scheduler ();
 	}
-	
 	atomic_set (&tx_pending, 0);
-} /* tx_dpc */
+} /* tx_handler */
 
 /*---------------------------------------------------------------------------*\
 \*---------------------------------------------------------------------------*/
-static void rx_dpc (void * data) {
+static void tx_task (unsigned long data) {
+	
+	UNUSED_ARG (data);
+	if (in_critical ()) {
+		atomic_set (&tx_flag, 1);
+		kick_scheduler ();
+	} else if (spin_trylock (&stack_lock)) {
+		tx_handler (capi_card);
+		spin_unlock (&stack_lock);
+	} else {
+		atomic_set (&tx_flag, 1);
+		kick_scheduler ();
+	}
+} /* tx_task */
+
+/*---------------------------------------------------------------------------*\
+\*---------------------------------------------------------------------------*/
+static void rx_handler (card_p cp) {
 	int			is_open;
-	card_p			cp;
 	completer_func_t	fp;
 	void *			bufp;
 	xfer_info_t		info;
@@ -1808,23 +1828,20 @@ static void rx_dpc (void * data) {
 	unsigned		deltaT;
 #endif
 	
+	atomic_set (&rx_flag, 0);
 	is_open = atomic_read (&link_open);
 	info (is_open);
 	if (!is_open) {
 		return;
 	}
-	
 	pending = atomic_xchg (&rx_pending, 1);
 	assert (pending == 0);
-
-	cp = (card_p) data;
-	assert (cp);
+	assert (cp != NULL);
 	assert (PEEK (cp->io_base + XFER_TOPC_STATUS) == PC_PENDING);
 	xfer_info_get (rx_info, 0, &info);
 	assert (info.state == TM_READY);
 	assert (((info.length + 3) & ~3) <= DMA_BUFFER_SIZE);
 	assert (info.length < cp->rx_length);
-	
 	assert (cp->rx_length > 0);
 	assert (cp->rx_func != NULL);
 	assert (cp->rx_buffer != NULL);
@@ -1842,9 +1859,25 @@ static void rx_dpc (void * data) {
 		(*fp) (bufp, info.length, cp->rx_ctx);
 		kick_scheduler ();
 	}
-
 	atomic_set (&rx_pending, 0);
-} /* rx_dpc */
+} /* rx_handler */
+
+/*---------------------------------------------------------------------------*\
+\*---------------------------------------------------------------------------*/
+static void rx_task (unsigned long data) {
+	
+	UNUSED_ARG (data);
+	if (in_critical ()) {
+		atomic_set (&rx_flag, 1);
+		kick_scheduler ();
+	} else if (spin_trylock (&stack_lock)) {
+		rx_handler (capi_card);
+		spin_unlock (&stack_lock);
+	} else {
+		atomic_set (&rx_flag, 1);
+		kick_scheduler ();
+	}
+} /* rx_task */
 
 /*---------------------------------------------------------------------------*\
 \*---------------------------------------------------------------------------*/
@@ -1872,28 +1905,21 @@ static void irq_handler (int irq, void * args, struct pt_regs * regs) {
 		POKE (capi_card->io_base + INT_CTL, flags);
 		assert ((PEEK (capi_card->io_base + INT_CTL) 
 				& ~(CARD_PCI_INT_ASSERT | CARD_PCI_INT_ISASSERTED)) != 0);
-
 		if (!atomic_read (&link_open)) {
 			return;
 		}
-
 		tx_flag = PEEK (capi_card->io_base + XFER_TOTM_STATUS) == TM_READY;
 		rx_flag = PEEK (capi_card->io_base + XFER_TOPC_STATUS) == TM_READY;
 		if (tx_flag) {
 			POKE (capi_card->io_base + XFER_TOTM_STATUS, PC_PENDING); 
-			queue_task (&tq_tx_dpc, &tq_private); 
+			tasklet_schedule (&tx_tasklet);
 		}
 		if (rx_flag) {
 			POKE (capi_card->io_base + XFER_TOPC_STATUS, PC_PENDING);
-			queue_task (&tq_rx_dpc, &tq_private);
+			tasklet_schedule (&rx_tasklet);
 		}
 	}
-
 	info (0 == (PEEK (capi_card->io_base + INT_CTL) & CARD_PCI_INT_ASSERT));
-	if (tx_flag || rx_flag) {
-		queue_task (&tq_dpc, &tq_immediate); 
-		mark_bh (IMMEDIATE_BH);	
-	}
 } /* irq_handler */
 
 /*---------------------------------------------------------------------------*\
@@ -1958,25 +1984,6 @@ static void kick_scheduler (void) {
 	wake_up_interruptible (&wait);
 } /* kick_scheduler */
 
-/*---------------------------------------------------------------------------*\
-\*---------------------------------------------------------------------------*/ 
-static void scheduler_control (unsigned ena) {
-	int	enabled = (int) ena;
-	int	changed;
-
-	enter_critical ();
-	changed = (atomic_read (&scheduler_enabled) != enabled);
-	leave_critical ();
-	if (changed) {
-		if (enabled) {
-			kick_scheduler ();
-		} else {
-			atomic_set (&scheduler_enabled, 0);
-			wake_up_interruptible (&wait);
-		}
-	}
-} /* scheduler_control */
-
 /*---------------------------------------------------------------------------*\
 \*---------------------------------------------------------------------------*/ 
 static int trigger_tm_interrupt (void) {
