@@ -103,6 +103,10 @@ extern struct capi_ctr *		capi_controller[2];
 
 /*---------------------------------------------------------------------------*\
 \*---------------------------------------------------------------------------*/
+#if defined (TM_TIMESTAMP)
+extern unsigned get_tm_timer (void);
+#endif
+
 extern void * deref_map_addr_from_tm_ptr (card_p cp, unsigned long X);
 extern void * map_addr_from_tm_ptr (card_p cp, unsigned long X);
 	
