@@ -280,7 +280,7 @@ int check_timer_cb (dbg_timer * t, void (* callback) (dbg_timer *, struct timeva
 	|| ((delta.tv_sec == t->d.tv_sec) && (delta.tv_usec > t->d.tv_usec))
 	) {
 		if (callback != NULL) {
-			(*callback) (t, delta);
+			(*callback) (t, &delta);
 		}
 		res = 0;
 	} 
